/***************************************************************************
 *                                                                         *
 *    NoMoRe++                                                             *
 *                                                                         *
 *    Copyright (C) 2003-2005 NoMoRe Developing Group                      *
 *                                                                         * 
 *    For more information, see http://www.cs.uni-potsdam.de/nomore/       *
 *    or email to nomore-dg@cs.uni-potsdam.de                              *
 *                                                                         *
 *                                                                         *
 *    This program is free software; you can redistribute it and/or        *
 *    modify it under the terms of the GNU General Public License          *
 *    as published by the Free Software Foundation; either version 2       *
 *    of the License, or (at your option) any later version.               *
 *                                                                         *
 *    This program is distributed in the hope that it will be useful,      *
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 *    GNU General Public License for more details.                         *
 *                                                                         *
 *    You should have received a copy of the GNU General Public License    *
 *                                                                         * 
 ***************************************************************************/

#if defined (_MSC_VER) && _MSC_VER <= 1300
#pragma warning (disable : 4786)
#endif

#include "nomore_app.h"
#include "nomore.h"
#ifdef KDEVELOP  
  #include "value.h"  
#else
  #include "program_opts/value.h"  
#endif

#include <graph.h>
#include <lparsereader.h>
#include <object_factory.h>
#include <choice_operator.h>
#include <operators/backward_prop.h>
#include <operators/lookahead.h>
#include <heuristics/neighborhood.h>
#include <printer/null_printer.h>
#include <printer/stdout_printer.h>
#include <stats.h>
#include <statistics/mode_watcher.h>
#include <statistics/timing_statistics.h>
#include <statistics/operator_call_statistics.h>
#include <statistics/standard_statistics.h>
#include <nary_backtracking_solver.h>
#include <binary_backtracking_solver.h>
#include <string>
#include <iostream>
#include <istream>
#include <fstream>
#include <ostream>
#include <cctype>
#include <iomanip>
using namespace ProgramOptions;
using namespace std;
using namespace NS_NOMORE;

namespace {

const char DEFAULT_OPS[]        = "PBL";
const char DEFAULT_CHOICE[]     = "C";
const char DEFAULT_HEURISTIC[]  = "Neighborhood";

void errorExit(const UnknownObject& u) {
  string typeAsString;
  switch(u.getObjectType()) {
  case ObjectFactory::DET_OP:
    typeAsString = "propagation-operator";
    break;
  case ObjectFactory::LOOKAHEAD_OP:
    typeAsString = "lookahead-operator";
    break;
  case ObjectFactory::CHOICE_OP:
    typeAsString = "choice-operator";
    break;
  case  ObjectFactory::HEURISTIC:
    typeAsString = "heuristic";
    break;
  default:
    typeAsString = "object";
  }
  string msg = string("'") + u.what() + "' unknown " + typeAsString + "!\n";
  if (u.getObjectType() != ObjectFactory::HEURISTIC) {
    msg += "use --list-operators to get a list of all known operators.";
  }
  else {
    msg += "use --list-heuristics to get a list of all known heuristics.";
  }
  throw CommandLineError(msg);
}


typedef std::map<std::string, std::string> OperatorMap;
OperatorMap initOperatorMap() {
  // P is always required
  OperatorMap ret;
  ret["P"]    = "P";
  ret["B"]    = ret["BP"] = "(P,B)*";
  ret["U"]    = ret["PU"] = "(P,U)*";
  ret["L"]    = ret["LP"] = ret["LU"] = "(P,L)*";
  ret["BU"]   = ret["BPU"] ="((P,B)*,U)*";
  ret["BL"]   = ret["BLP"] = ret["BLU"] = ret["BLUP"] = "((P,B)*,L)*";
  return ret;
}

OperatorMap operatorMap_g = initOperatorMap();
bool isAlpha(char c) {
  return isalpha(static_cast<unsigned char>(c)) != 0;
}
std::string lookup(const string& opStr) {
  std::string alphaStr;
  for (std::string::size_type i = 0; i != opStr.size(); ++i)
    if (isAlpha(opStr[i]))
      alphaStr += opStr[i];
  
  sort(alphaStr.begin(), alphaStr.end());
  OperatorMap::iterator it = operatorMap_g.find(alphaStr);
  if (it != operatorMap_g.end()) {
    return it->second;
  }
  errorExit(UnknownObject(opStr, ObjectFactory::DET_OP));
  return "";
}
 
istream& getInputStream(const string& fn) {
  static ifstream file;
  if (!fn.empty()) {
    file.open(fn.c_str());
    if (!file.is_open())
      throw CommandLineError(fn + " no such file");
    return file;
  }
  return cin;
}


}
NomoreApp::NomoreApp(int& argc, char** argv) 
  : options_((string("nomore++ ") + string(VERSION) + 
              string("\nusage: nomore [number] [options]\n")).c_str())
  , factory_(new ObjectFactory) {
  
  factory_->registerKnownClasses();
  
  OptionGroup generalOps("general options:");
  generalOps.addOptions()
    ("file,f", value<string>()->defaultValue(""), "Read from file arg instead of from stdin")
    ("version,V",	bool_switch(),	"Print version and exit")
		("help,h",		bool_switch(),	"Print help and exit")
    ("n-ary-backtracking,n", bool_switch(), "Use n-ary backtracking solver")
    ("force-tightness-check", bool_switch(), "Check program tightness")
    ("tightness-gradient,t", value<int>()->defaultValue(3), "Gradient for preferring nodes (0 to disable tightness-factor computation)")
  ;

  OptionGroup opOps("operator options:");
  opOps.addOptions()     
    ("choice-op", value<string>()->defaultValue(DEFAULT_CHOICE), "Choice operator to use")
    ("prop-op", value<string>()->defaultValue(DEFAULT_OPS), std::string("List of propagation operators to use (default: ").append(DEFAULT_OPS).append(")").c_str())
    ("lookahead-op", value<string>(), string("Propagation operators to use in lookahead (default: ").append(DEFAULT_OPS).append(")").c_str())
    ("list-operators", bool_switch(), "Display a short description of all known operators")
    ("heuristic", value<string>(), "Heuristic to use")
    ("neighborhood-level,l", value<int>(), "Iterations to use in Neighborhood-Heuristic")
    ("list-heuristics", bool_switch(), "Display a short description of all known heuristics")
    ("no-heuristic", bool_switch(), "Do not use a heuristic")
    ("no-lookahead", bool_switch(), "Do not use lookahead")
    ("body-lookahead", bool_switch(), "Use body lookahead instead of hybrid lookahead")
    ("no-jumping", bool_switch(), "Disbale jumping in B-Operator")
  ;

  OptionGroup outOps("output options:");
  outOps.addOptions()
    ("info,i", bool_switch(), "displays the values of selected options")
    ("no-output", bool_switch(), "do not print answer sets")
    ("no-statistics", bool_switch(), "do not print statistics")
    ("timings", bool_switch(), "print timing statistics")
    ("operator-calls", bool_switch(), "print operator call statistics")
    ("full-statistics", bool_switch(), "print all statistics")
  ;
  
  options_.addOptions(generalOps);
  options_.addOptions(opOps);
  options_.addOptions(outOps);

  OptionGroup allOpts;
  allOpts.addOptions(options_);
  allOpts.addOptions()
    ("number", value<int>()->defaultValue(1))
    ("operators,o", value<string>(), "op-string with format: ChoiceOp:DetOp:PostOp:LOp[DetOp]:Heu")
  ;

  try {
    values_.store(parseCommandLine(argc, argv, allOpts, false, "number"));
  }
  catch(const std::exception& e) {
    throw CommandLineError(e.what());
  } 

}

void NomoreApp::printHelp() const {
  cout << options_ << endl;
	cout << "Number is the number of answer sets to compute. A zero indicates all." << endl;
}

void NomoreApp::printVersion() const {
  cout << "NoMore++ - Version: " << VERSION << endl;
  cout << "Copyright (C) 2003-2005 NoMoRe Developing Group" << endl;
  cout << "This software is provided \"as is\" without expressed or implied warranty,\n"
       << "and with no claim as to its suitability for any purpose." << endl;
}

void NomoreApp::printOperators() const {
  cout << "NoMore++ - Version: " << VERSION << " supports the following operators: " << endl;
  vector<string> ops = factory_->getRegisteredClasses(ObjectFactory::DET_OP);
  cout << "Propagation-Operators:\n"
       << "----------------------\n";
  for (size_t i = 0; i < ops.size(); ++i) {
    std::string d = factory_->getDescription(ops[i]);
    if (!d.empty())
      cout << ops[i] << "\t" << d << "\n";
  }
  cout << "\n";
  ops = factory_->getRegisteredClasses(ObjectFactory::LOOKAHEAD_OP);
  cout << "Lookahead-Operators:\n"
       << "--------------------\n";

  for (size_t i = 0; i < ops.size(); ++i) {
    cout << ops[i] << "\t" << factory_->getDescription(ops[i]) << "\n";
  }
  cout << "\n";
  ops = factory_->getRegisteredClasses(ObjectFactory::CHOICE_OP);
  cout << "Choice-Operators:\n"
       << "-----------------\n";
  
  for (size_t i = 0; i < ops.size(); ++i) {
    cout << ops[i] << "\t" << factory_->getDescription(ops[i]) << "\n";
  }
  cout << endl;

}

void NomoreApp::printHeuristics() const {
  cout << "NoMore++ - Version: " << VERSION << " supports the following heuristics: " << endl;
  cout << "Note:\n"
       << "\tHeuristics rate possible choice points.\n"
       << "\tDynamic Heuristics reevaluate candidates whenever a choice is made.\n"
       << "\tStatic Heuristics rate candidates only once at program start.\n"
       << "\tLookahead Heuristics rate candidates after every lookahead operation.\n\n";
  vector<string> heus = factory_->getRegisteredClasses(ObjectFactory::HEURISTIC);
  cout << left << setw(10) << "Name" << "  " << "Description\n";
  cout << "-----------------------" << endl;
  for (size_t i = 0; i < heus.size(); ++i) {
    cout << left << setw(10) << heus[i] << "  " << factory_->getDescription(heus[i]) << "\n";
  }
  cout << endl;
}

int NomoreApp::run() {
  if (values_.count("help")) {
    printHelp();
    return 0;
  }
  if (values_.count("version")) {
    printVersion();
    return 0;
  }
  if (values_.count("list-operators")) {
    printOperators();
    return 0;
  }
  if (values_.count("list-heuristics")) {
    printHeuristics();
    return 0;
  }
  validateArguments();
  
  std::istream& input = getInputStream(value_cast<string>(values_["file"]));
  Graph g(value_cast<int>(values_["tightness-gradient"]));
  if (values_.count("force-tightness-check")) {
    g.checkTightness();
  }
  std::auto_ptr<Statistics> stats = createStatisticObject(g);
  std::auto_ptr<Solver> solver(createSolver(g, *stats, buildOpString()));
  if (values_.count("neighborhood-level")) {
    if (Neighborhood* h = dynamic_cast<Neighborhood*>(&solver->getChoiceOperator()->getHeuristic())) {
      h->setMaxDepth(value_cast<int>(values_["neighborhood-level"]));
    }
  }
  LparseReader r;
  cout << "Reading program..." << flush;
  stats->startTimer();
  r.readProgram(input, g);
  cout << "done" << endl;
  std::cout << "solve ... " << std::endl;  
  if(solver->solve(g))
    std::cout << "YES" << endl;
  else
    std::cout << "NO" << endl;
  stats->stopTimer();
  cout << "Answer sets found: " << solver->getNrOfFoundAS() << endl;
  stats->print(std::cout);

  Graph::Tightness t = g.tightness();
  cout << "Program-Tightness: " << (t == Graph::tight ? "tight" : t == Graph::non_tight ? "non-tight" : "indeterminate") << endl;
  cout << "Rules: " << g.countRules() << "\n";
  cout << "Head-Nodes: " << g.countHeadNodes() << "\n";
  cout << "Body-Nodes: " << g.countBodyNodes() << endl;

  if (values_.count("info") != 0)
    printInfo(*solver);
  
  
  return 0;
}

void NomoreApp::validateArguments() const {
  if (value_cast<int>(values_["number"]) < 0)
    throw CommandLineError("number can not be negative.");
  if (value_cast<int>(values_["tightness-gradient"]) < 0)
    throw CommandLineError("'tightness-gradient' can not be negative.");
  
  if (values_.count("neighborhood-level") && value_cast<int>(values_["neighborhood-level"]) < 1)
    throw CommandLineError("'neighborhood-level' must be greater than 0!");

  if (values_.count("full-statistics") && values_.count("no-statistics"))
    throw CommandLineError("options 'full-statistics' and 'no-statistics' are in conflict!");

  if (values_.count("no-lookahead") 
    && (values_.count("lookahead-op") || values_.count("body-lookahead"))) {
    throw CommandLineError("option 'no-lookahead' can't be used together with 'lookahead-op' or 'body-lookahead'!");
  }

  if (values_.count("no-heuristic") && values_.count("heuristic")) {
    throw CommandLineError("option 'no-heuristic' can't be used together with 'heuristic'!");
  }

  if (values_.count("heuristic") && value_cast<string>(values_["heuristic"]) != "Neighborhood"
    && values_.count("neighborhood-level")) {
    cerr << "Warning: Option 'neighborhood-level' must be used with heuristic 'Neighborhood'. Option ignored!" << endl;
  }
}

string NomoreApp::buildOpString() const {
  if (values_.count("operators")) {
    return value_cast<string>(values_["operators"]);
  }

  std::string opString = value_cast<string>(values_["choice-op"]);
  opString += ":" + lookup(value_cast<string>(values_["prop-op"]));
  if (values_.count("no-lookahead") != 0) {
    opString += ":None:";
  }
  else {
    std::string lop = values_.count("body-lookahead") ? ":BLA" : ":HLA";
    if (values_.count("lookahead-op"))
      lop += "[" + lookup(value_cast<string>(values_["lookahead-op"])) + "]:";
    else
      lop += ":";
    opString += lop;
  }
  return addHeuristicName(opString);
}

std::auto_ptr<Solver> NomoreApp::createSolver(NS_NOMORE::Graph& g, NS_NOMORE::Statistics& stats,
                               const std::string& opString) const {
  std::auto_ptr<Solver> p(values_.count("n-ary-backtracking") ? static_cast<Solver*>(new NaryBacktrackingSolver) : static_cast<Solver*>(new BinaryBacktrackingSolver));
  p->setNrOfAnswerSetsToCompute(value_cast<int>(values_["number"]));
  p->setAnswerSetPrinter(
    values_.count("no-output") 
      ? static_cast<Printer*>(new NullPrinter()) 
      : new StdoutPrinter()
  );

  factory_->setOperatorDecoratorFactory(stats);
  try {
    ObjectFactory::Operators o = factory_->createOperators(g, opString, true);
    bool nj = values_.count("no-jumping") != 0;
    if ( nj  ) {
      if (BackwardPropagator* b = static_cast<BackwardPropagator*>(o.detOp_->extractOperator("B"))) {
        if (nj) b->disableJumping();
      }
      else {
        cerr << "Warning: Option 'no-jumping' is superfluous. B-Operator is not active!" << endl;
      }
    }
    p->setPropagationOperator(o.detOp_.release(), o.lookOp_.release());
    p->setChoiceOperator(o.choiceOp_.release());
  }
  catch(const UnknownObject& e) {    
    errorExit(e);
  }
  catch (const FormatError& e) {    
    throw CommandLineError(string("Option 'operators': ").append(e.what()));
  }
  catch (const InvalidOperatorCombination& e) {    
    throw CommandLineError(string("Option 'operators': ").append(e.what()));
  }
  return p;
}


std::auto_ptr<Statistics> NomoreApp::createStatisticObject(Graph& g) const {
  std::auto_ptr<Statistics> s(new Statistics());

  if (values_.count("no-statistics"))
    return s;

  s->add(new StandardStatistics(g.getEventManager()));
  s->add(new ModeWatcher(g.getEventManager()));

  if (values_.count("timings") || values_.count("full-statistics")) {
    s->add(new TimingStatistics(g.getEventManager()));
  }
    
  if(values_.count("operator-calls") || values_.count("full-statistics")) {
    OperatorCalls *opCalls = new OperatorCalls(g.getEventManager());
    s->add(opCalls);
    s->setDecorator(opCalls);
  }

  return s;
}

void NomoreApp::printInfo(const Solver& p) const {
  std::pair<Operator*, Operator*> det = p.getPropagationOperator();
  if (values_.count("operators") != 0) {
    cout 
      << "Operators                     : " << value_cast<std::string>(values_["operators"]) << "\n";
  } 
  else {
    cout 
      << "Propagation-Operators         : "  << det.first->getName() << "\n"
      << "Choice-Operator               : " << p.getChoiceOperator()->getName() << "\n"
      << "Heuristic                     : " << p.getChoiceOperator()->getHeuristic().getName() << "\n";
    std::string lop = det.second->getName();
    if (Lookahead* lookahead = dynamic_cast<Lookahead*>(det.second)) {
      cout
        << "Lookahead                     : " << lookahead->getName() << "\n"
        << "Lookahead-Propagation         : " << lookahead->getPropagationOperator().getName() << "\n";
    }
    else {
      cout
        << "Lookahead                     : disabled\n";
    }
    cout
      << "Backpropagtion                : " << (det.first->extractOperator("B") != 0 ? "Enabled" : "Disabled") << "\n"
      << "U-Operator                    : " << (det.first->extractOperator("U") != 0 ? "Enabled" : "Disabled") << "\n";  
  }
  cout << "Number of answer sets to find : ";
  if (p.getNrOfAnswerSetsToCompute() == -1)
    cout << "all";
  else
    cout << p.getNrOfAnswerSetsToCompute();
  cout << endl;
}

std::string NomoreApp::addHeuristicName(std::string opStr) const {
  if (values_.count("no-heuristic") != 0) {
    opStr += "None";
  }
  else if (values_.count("heuristic") != 0) {
    std::string heu = value_cast<std::string>(values_["heuristic"]);
    if ( (heu == "BodyLA" && !values_.count("body-lookahead"))
      || (heu == "HybridLA" && values_.count("hybrid-lookahead"))
      || (values_.count("no-lookahead") != 0 && (heu == "BodyLA" || heu == "HybridLA"))) {
      cerr << "warning: Heuristic is incompatible with lookahead. Heuristic is ignored." << endl;
      heu = "None";
    }
    opStr += heu;
  }
  else if (values_.count("no-lookahead") == 0) {
    opStr += values_.count("body-lookahead") ? "BodyLA" : "HybridLA";
  }
  else {
    opStr += DEFAULT_HEURISTIC;
  }
  return opStr;
}
